@@ -247,6 +247,8 @@ public class ControlWindow extends AuxWindow
         super( "TOPCAT", null );
 
         /* Configure table factory. */
+        tabfact_.setPreparation( new TopcatPreparation( tabfact_
+                                                       .getPreparation() ) );
         tabfact_.getJDBCHandler()
                 .setAuthenticator( new TextModelsAuthenticator() );
         taboutput_.setJDBCHandler( tabfact_.getJDBCHandler() );
@@ -819,7 +821,7 @@ public class ControlWindow extends AuxWindow
     public TopcatModel addTable( StarTable table, String location,
                                  boolean select ) {
         TopcatModel tcModel =
-            TopcatCodec.getInstance().decode( table, location, this );
+            TopcatUtils.decodeSession( table, location, this );
         if ( tcModel == null ) {
             tcModel = TopcatModel
                      .createDefaultTopcatModel( table, location, this );
